#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<math.h>
#include"myConst.h"
#include"fold.h"
#include"energy_const.h"
#include"fold_vars.h"
#include"pair_mat.h"
#include"convert_Vienna.h"
#include"params.h"
#include"ParCal.h"
#include<limits.h>
#include"misc.h"
#include"RNAmultiloop.h"

/*Partition function calculation for the RNAmultiloop program*/
/* flag=1-> multiloop number partition function, flag=2-> multiloop order partition function*/

int GetZB(double ***ZB, double ***ZM, double ***ZM1, int i, int j, int m, char sequence[MAXSIZE], int flag){
  int l,r,k;
  if(m==0)
    {ZB[i][j][m]=exp(-HP_Energy(i,j,S0,sequence+1)/kT);
      ZB[j][i][m]=1;
      for (l=i+1;l<min(i+30,j-5)+1;++l)
        {for(r=max(l+4,j-(30-(l-i)));r<j;++r)
            {if(BP(l,r,sequence))
               {ZB[i][j][m]+=ZB[l][r][m]*exp(-IL_Energy(i,j,l,r,S0)/kT);
                 ZB[j][i][m]+=ZB[r][l][m];
               }
            }
        }
    }
  else
    {
      if(flag==1)
	  {
	    for (l=i+1;l<min(i+30,j-5)+1;++l)
        	{for(r=max(l+4,j-(30-(l-i)));r<j;++r)
            	    {if(BP(l,r,sequence))
               		{ZB[i][j][m]+=exp(-IL_Energy(i,j,l,r,S0)/kT)*ZB[l][r][m];
                 	 ZB[j][i][m]+=ZB[r][l][m];
               		}
            	    }
        	}
      	    for(r=i+6;r<j-4;++r)
                {for(k=0;k<m;++k)
            	    {ZB[i][j][m]+=exp(-(ML_close+MLbasepairAndAUpenalty(j,i,S0))/kT)*ZM[i+1][r-1][k]*ZM1[r][j-1][m-k-1];
              	     ZB[j][i][m]+=ZM[r-1][i+1][k]*ZM1[j-1][r][m-k-1];
                    }
                }
	  }
	if(flag==2)
	  {
	    for (l=i+1;l<min(i+30,j-5)+1;++l)
        	{for(r=max(l+4,j-(30-(l-i)));r<j;++r)
            	    {if(BP(l,r,sequence))
               		{ZB[i][j][m]+=exp(-IL_Energy(i,j,l,r,S0)/kT)*ZB[l][r][m];
                 	 ZB[j][i][m]+=ZB[r][l][m];
               		}
            	    }
        	}
      	    for(r=i+6;r<j-4;++r)
        	{k=m-1;
          	 for(l=0;l<m;++l)
            		{ZB[i][j][m]+=exp(-(ML_close+MLbasepairAndAUpenalty(j,i,S0))/kT)*ZM[i+1][r-1][k]*ZM1[r][j-1][l];
              		ZB[j][i][m]+=ZM[r-1][i+1][k]*ZM1[j-1][r][l];
            		}
        	}
      	    for(r=i+6;r<j-4;++r)
        	{l=m-1;
          	 for(k=0;k<m-1;++k)
            		{ZB[i][j][m]+=exp(-(ML_close+MLbasepairAndAUpenalty(j,i,S0))/kT)*ZM[i+1][r-1][k]*ZM1[r][j-1][l];
              		ZB[j][i][m]+=ZM[r-1][i+1][k]*ZM1[j-1][r][l];
            		}
        	}
	  }
    }
}

int GetZM1(double ***ZB, double ***ZM, double ***ZM1, int i, int j, int m, char sequence[MAXSIZE], int flag)
{ int r;  // flag is not used in this function, since the M1 recursions are the same in multiloop number/order recursions.
  for (r=i+4;r<j+1;++r)
    { ZM1[i][j][m]+=ZB[i][r][m]*exp(-(ML_base*(j-r)+MLbasepairAndAUpenalty(i,r,S0))/kT);
      ZM1[j][i][m]+=ZB[r][i][m];
    }
}

int GetZM(double ***ZM, double ***ZM1, int i, int j, int m, char sequence[MAXSIZE], int flag)
{int r,k,l;
 for(r=i;r<j-3;++r)
    {ZM[i][j][m]+=ZM1[r][j][m]*exp(-ML_base*(r-i)/kT);
      ZM[j][i][m]+=ZM1[j][r][m];}
  if(flag==1) //multiloop number
    {
      for(r=i+5;r<j-3;++r)
      	{ for(k=0;k<m+1;++k)
        	{
		ZM[i][j][m]+=ZM[i][r-1][k]*ZM1[r][j][m-k];
          	ZM[j][i][m]+=ZM[r-1][i][k]*ZM1[j][r][m-k];
        	}	
    	}	
    }
  if(flag==2)  //multiloop order
    {
      for(r=i+5;r<j-3;++r)
    	{k=m;
      	 for(l=0;l<m+1;++l)
         	{ZM[i][j][m]+=ZM[i][r-1][k]*ZM1[r][j][l];
          	ZM[j][i][m]+=ZM[r-1][i][k]*ZM1[j][r][l];
        	}
    	}	
      for(r=i+5;r<j-3;++r)
    	{ l=m;
      	  for(k=0;k<m;++k) // to prevent k=m,l=m to be double-counted.
        	{ZM[i][j][m]+=ZM[i][r-1][k]*ZM1[r][j][l];
          	ZM[j][i][m]+=ZM[r-1][i][k]*ZM1[j][r][l];
        	}
    	}
    }
}

int GetZ(int j, int m, double ***ZB,double **Z, double **N, char sequence[MAXSIZE], int flag)
{int r,k,l;
 if(j<5)
   {if (m==0)
       {Z[j][m]=1;
         N[j][m]=1;
       }
     else
       {Z[j][m]=0;    // the empty structure is no longer counted in when m!=0 since it doesn't contain multiloops.
         N[j][m]=0;
       }
    }  
  else
   {
     if(flag==1)
	{
	  Z[j][m]+=Z[j-1][m];
      	  N[j][m]+=N[j-1][m];
      	  for(r=1;r<j-3;++r)
            {if(BP(r,j,sequence))
              {if(r==1)
                {Z[j][m]+=ZB[r][j][m]*exp(-AU_Penalty(1,j,S0)/kT);
                  N[j][m]+=ZB[j][r][m];}
               else
                {for(k=0;k<m+1;++k)
                    {Z[j][m]+=Z[r-1][k]*ZB[r][j][m-k]*exp(-AU_Penalty(r,j,S0)/kT);
                      N[j][m]+=N[r-1][k]*ZB[j][r][m-k];
                    }
                 }
               } 
             }   
	  }
     if(flag==2)
	{
	  Z[j][m]+=Z[j-1][m];
      	  N[j][m]+=N[j-1][m];
      	  for(r=1;r<j-3;++r)
            {if(BP(r,j,sequence))
              {if(r==1)
                {Z[j][m]+=ZB[r][j][m]*exp(-AU_Penalty(1,j,S0)/kT);
                  N[j][m]+=ZB[j][r][m];}
               else                
		{
                  {k=m;
                    for(l=0;l<m+1;++l)
                      {Z[j][m]+=Z[r-1][k]*ZB[r][j][l]*exp(-AU_Penalty(r,j,S0)/kT);
                        N[j][m]+=N[r-1][k]*ZB[j][r][l];
                      }
                   }
                   {l=m;
                     for(k=0;k<m;++k)
                      {Z[j][m]+=Z[r-1][k]*ZB[r][j][l]*exp(-AU_Penalty(r,j,S0)/kT);
                        N[j][m]+=N[r-1][k]*ZB[j][r][l];
                      }
                   }
                 }
               }
             }   
	 }
     }  
}



