#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<math.h> 
#include"fold.h"
#include"energy_const.h"
#include"fold_vars.h"
#include"pair_mat.h"
#include"convert_Vienna.h"
#include"params.h"
#include"myConst.h"
#include"misc.h"
#include"RNAmultiloop.h"
#include"McCaskill.h"
#include"limits.h"

PRIVATE short *encode_seq(const char *seq);
double MultiloopPartition(double *MP,double *MN, int M_MAX, char sequence[MAXSIZE], int flag);
/*This is an algorithm to calculate the partition function with respect to the number of multiloops or multiloop order under Turner Energy model.*/
/*Author: Yang Ding; Date: 07/04/2009*/

int main(int argc, char *argv[]){
  char sequence[MAXSIZE],filename[FILENAMELENGTH];
  FILE *inputfile;
  int i,j,h;
  double MP[MAXSIZE],MN[MAXSIZE];// MP is the array to store the values Multiloop Partition with k multiloops/ order-k multiloops, MN is a counting version of that.
  double totalpar; //total partition function
  int flag=0; // flag=1-> multiloop number partition function, flag=2-> multiloop order partition function
  int M_MAX=5; //default threshold: multiloop number/order =5
  if (argc==4||argc==6)
    {
      i=1;
      while(i<argc)
	{
	  if(!strcmp(argv[i], "-s"))
	    {
	      sequence[0]='@';
	      strncpy(sequence+1,argv[i+1],strlen(argv[i+1])); // the sequence starts from index 1
	      sequence[strlen(argv[i+1])+1]='\0'; //remember to end the sequence
	      i+=2;
	    }
	  else if(!strcmp(argv[i], "-n"))
	    {
	      M_MAX=atoi(argv[i+1]);
	      i+=2;
	    }
	  else if(!strcmp(argv[i], "-order"))
	    {
	     flag=2;
	     i+=1;
	    }
	  else if(!strcmp(argv[i], "-number"))
	    {
	     flag=1;
	     i+=1;
	    }
	  else
	    {
	      printf("Error: Unrecognizable Flag!\n");
	      printf("Usage:");
	      printf("%s -s SEQUENCE -number/-order [-n NUM(default 5)] \n",argv[0]);
	      printf("SEQUENCE is the RNA sequence.\n");
	      printf("NUM is the maximum number/order of  multiloop that one considers.\n");
              printf("-number/-order flag tells the program whether to calculate parametric partition function with respect to multiloop number or multiloop order\n");
	      printf("The length of RNA sequence should be less than %d and larger than or equal to 5.\n",MAXSIZE);
	      exit(1);
	    }
	}
    CheckSequence(sequence);
    S0=encode_seq(sequence+1);
    seqlen=strlen(sequence)-1;
    Initialize_Params();
    make_pair_matrix();//needed for pair matching
    kT = (temperature+K0)*GASCONST/1000.0;
    //printf("%.15f\n",kT);
    ML_base=(double)P->MLbase/100;
    ML_close=(double)P->MLclosing/100;
    for (i=0;i<M_MAX+1;++i)
      {MP[i]=0;
       MN[i]=0;}
    MultiloopPartition(MP,MN,M_MAX,sequence,flag);
    double** McCaskillZ;
    McCaskillZ=runMcCaskill(sequence);
    totalpar=McCaskillZ[1][seqlen];
    printf("The RNA sequence is %s:\n",sequence+1);
    if(flag==1)
	{
	printf("The first column is the number of multiloops in a secondary structure.\n");
	printf("The partition function and number of secondary structure with k-multiloop is:\n");
	}
    else
	{
	printf("The first column is the order of multiloops in a secondary structure.\n");
    	printf("The partition function and number of secondary structure with order-k-multiloop is:\n");
	}
    printf("%-20s%-30s%-30s%-30s\n","K-hairpin","Number","Partition Function","Probability");
    for (h=0;h<M_MAX+1;++h)
      printf("%-8d\t%-30.15lf\t%-30.15lf\t%-30.15lf\n",h,MN[h],MP[h],MP[h]/totalpar);
    }
  else{
    printf("Usage:");
    printf("%s -s SEQUENCE -number/-order [-n NUM(default 5)] \n",argv[0]);
    printf("SEQUENCE is the RNA sequence\n");
    printf("NUM is the maximum number/order of  multiloop that one considers.\n");
    printf("-number/-order flag tells the program whether to calculate parametric partition function with respect to multiloop number or multiloop order\n");
    printf("The length of RNA sequence should be less than %d and larger than or equal to 5\n",MAXSIZE);
    exit(1);
  }
  return 0;
}

/*Below modified from fold.c*/
PRIVATE short *encode_seq(const char *seq) {
  unsigned int k,l;
  short *S0_out;
  int i;
  l = strlen(seq);
  S0_out = (short *) space(sizeof(short)*(l+2));
  S0_out[0]=l;
  for (k=1; k<=l; k++) { /* make numerical encoding of seq */
    S0_out[k]= (short) encode_char(toupper(seq[k-1]));
  }
  return S0_out;
}

/*The following function is the main recursion*/
double MultiloopPartition(double *MP,double *MN, int M_MAX, char sequence[MAXSIZE], int flag)
{ double ***ZB;
  double ***ZM;
  double ***ZM1;
  double **Z;
  double **N;
  Z=Allocate2DMatrix( seqlen+1, M_MAX+1);
  N=Allocate2DMatrix( seqlen+1, M_MAX+1);
  ZB=Allocate3DMatrix( seqlen+1,seqlen+1, M_MAX+1);
  ZM=Allocate3DMatrix( seqlen+1,seqlen+1, M_MAX+1);
  ZM1=Allocate3DMatrix( seqlen+1,seqlen+1, M_MAX+1);
  int i,j,r,d,m;
  for(i=1;i<seqlen+1;++i)
    {for (j=1;j<seqlen+1;++j)
	{for(r=1;r<M_MAX+1;++r)
	    { ZB[i][j][r]=0;
	      ZM[i][j][r]=0;
	      ZM1[i][j][r]=0;
	    }
	}
    }
  for (i=1;i<seqlen+1;++i)
    {for(j=1;j<M_MAX+1;++j)
	{Z[i][j]=0;
	  N[i][j]=0; 
	}
    }
  for(m=0;m<M_MAX+1;++m)
    {for(d=4;d<seqlen;++d)
	{for(i=1;i<seqlen-d+1;++i)
	    {j=i+d;
	     if(BP(i,j,sequence))
		GetZB(ZB,ZM,ZM1,i,j,m,sequence,flag); //flag is to specify whether one calculates partition function with respect to multiloop number of multiloop depth
	     GetZM1(ZB,ZM,ZM1,i,j,m,sequence,flag);
	     GetZM(ZM,ZM1,i,j,m,sequence,flag);
	     }
	  }
	  for(j=1;j<seqlen+1;++j)
	    {GetZ(j,m,ZB,Z,N,sequence,flag);}
	  MP[m]=Z[seqlen][m];
	  MN[m]=N[seqlen][m];
    }
}

